<?php
/**
 * SPDX-FileCopyrightText: 2024 Sascha Nitsch (grumpydeveloper) https://contentnation.net/@grumpydevelop
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * @author Yannis Vogel (vogeldevelopment)
 **/

namespace Federator\Api\FedUsers;

/**
 * handle activitypub following requests
 */
class Following implements \Federator\Api\FedUsers\FedUsersInterface
{
    /**
     * main instance
     *
     * @var \Federator\Api $main
     */
    private $main;

    /**
     * constructor
     * @param \Federator\Api $main main instance
     */
    public function __construct($main)
    {
        $this->main = $main;
    }

    /**
     * handle get call
     *
     * @param string|null $_user user to fetch followers for
     * @return string|false response
     */
    public function get($_user)
    {
        if (!isset($_user)) {
            return false;
        }
        $dbh = $this->main->getDatabase();
        $cache = $this->main->getCache();
        $connector = $this->main->getConnector();

        // get user
        $user = \Federator\DIO\User::getUserByName(
            $dbh,
            $_user,
            $connector,
            $cache
        );
        if ($user->id === null) {
            return false;
        }

        $following = new \Federator\Data\ActivityPub\Common\Following();
        $followingItems = \Federator\DIO\Followers::getFollowingForUser($dbh, $user->id, $connector, $cache);

        $config = $this->main->getConfig();
        $domain = $config['generic']['externaldomain'];
        $baseUrl = 'https://' . $domain . '/users/' . $_user . '/following';

        $pageSize = 10;
        $page = $this->main->extractFromURI("page", "");
        $id = $baseUrl;
        $items = [];
        $totalItems = count($followingItems);

        if ($page !== "") {
            $pageNum = max(0, (int) $page);
            $offset = (int) ($pageNum * $pageSize);
            $pagedItems = array_slice($followingItems, $offset, $pageSize);

            foreach ($pagedItems as $followed) {
                $items[] = $followed->actorURL;
            }
            $following->setItems($items);
            $id .= '?page=' . urlencode($page);
        }

        $following->setID($id);
        $following->setPartOf($baseUrl);
        $following->setTotalItems($totalItems);

        // Pagination navigation
        $lastPage = max(0, ceil($totalItems / $pageSize) - 1);

        if ($page === "" || $following->getCount() == 0) {
            $following->setFirst($baseUrl . '?page=0');
            $following->setLast($baseUrl . '?page=' . $lastPage);
        }
        if ($page !== "") {
            $pageNum = max(0, (int) $page);
            if ($pageNum < $lastPage) {
                $following->setNext($baseUrl . '?page=' . ($pageNum + 1));
            }
            if ($pageNum > 0) {
                $following->setPrev($baseUrl . '?page=' . ($pageNum - 1));
            }
            $following->setFirst($baseUrl . '?page=0');
            $following->setLast($baseUrl . '?page=' . $lastPage);
        } else {
            $following->setType('OrderedCollection');
        }
        $obj = $following->toObject();

        return json_encode($obj, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT);
    }

    /**
     * handle post call
     *
     * @param string|null $_user user to add data to outbox @unused-param
     * @return string|false response
     */
    public function post($_user)
    {
        return false;
    }
}
