<?php
/**
 * SPDX-FileCopyrightText: 2024 Sascha Nitsch (grumpydeveloper) https://contentnation.net/@grumpydevelop
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * @author Sascha Nitsch (grumpydeveloper)
 **/

namespace Federator\Data\ActivityPub\Common;

class OrderedCollection extends Collection
{
    /**
     * nested items
     *
     * @var APObject[]|string[]
     */
    protected $orderedItems = [];

    public function __construct()
    {
        parent::__construct();
        parent::addContext('https://www.w3.org/ns/activitystreams');
    }

    /**
     * convert internal state to php array
     *
     * @return array<string,mixed>
     */
    public function toObject()
    {
        $return = parent::toObject();
        $return['type'] = 'OrderedCollection';
        if (sizeof($this->orderedItems) > 0) {
            foreach ($this->orderedItems as $item) {
                if (is_string($item)) {
                    $return['orderedItems'][] = $item;
                } elseif (is_object($item)) {
                    $return['orderedItems'][] = $item->toObject();
                }
            }
        }
        return $return;
    }

    /**
     * create object from json
     *
     * @param array<string,mixed> $json input json
     * @return bool true on success
     */
    public function fromJson($json)
    {
        $success = parent::fromJson($json);
        if (!$success) {
            return false;
        }
        if (array_key_exists('orderedItems', $json)) {
            foreach ($json['orderedItems'] as $item) {
                $obj = \Federator\Data\ActivityPub\Factory::newActivityFromJson($item);
                if ($obj !== false) {
                    $this->orderedItems[] = $obj;
                }
            }
        }

        return true;
    }

    /**
     * add item to collection
     * @param APObject|string $item
     */
    public function append(&$item): void
    {
        $this->orderedItems[] = $item;
    }

    /**
     * get item with given index
     *
     * @return APObject|string|false
     */
    public function get(int $index)
    {
        if ($index >= 0) {
            if ($index >= sizeof($this->orderedItems)) {
                return false;
            }
            return $this->orderedItems[$index];
        } else {
            if (sizeof($this->orderedItems) + $index < 0) {
                return false;
            }
            return $this->orderedItems[sizeof($this->orderedItems) + $index];
        }
    }

    public function getCount(): int
    {
        return sizeof($this->orderedItems);
    }
    /**
     * set items
     *
     * @param APObject[]|string[] $items
     * @return void
     */
    public function setItems(&$items)
    {
        $this->orderedItems = $items;
    }
}
