<?php
/**
 * SPDX-FileCopyrightText: 2024 Sascha Nitsch (grumpydeveloper) https://contentnation.net/@grumpydevelop
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * @author Yannis Vogel (vogeldevelopment)
 **/

namespace Federator\DIO;

/**
 * IO functions related to votes
 */
class Votes
{
    /**
     * Add a vote (like/dislike)
     *
     * @param \mysqli $dbh
     * @param string $userId      The user who votes
     * @param string $targetId    The object being voted on (e.g., post id)
     * @param string $type        'like' or 'dislike'
     * @return string|false       The generated vote ID on success, false on failure
     */
    public static function addVote($dbh, $userId, $targetId, $type)
    {
        // Check if already voted
        $sql = 'SELECT id FROM votes WHERE user_id = ? AND target_id = ? AND type = ?';
        $stmt = $dbh->prepare($sql);
        if ($stmt === false) {
            throw new \Federator\Exceptions\ServerError();
        }
        $stmt->bind_param("sss", $userId, $targetId, $type);
        $foundId = 0;
        $ret = $stmt->bind_result($foundId);
        $stmt->execute();
        if ($ret) {
            $stmt->fetch();
        }
        $stmt->close();
        if ($foundId != 0) {
            return false; // Already voted
        }

        // Generate a unique ID for the vote
        do {
            $id = bin2hex(openssl_random_pseudo_bytes(16));
            // Check if the generated ID is unique
            $sql = 'SELECT id FROM votes WHERE id = ?';
            $stmt = $dbh->prepare($sql);
            if ($stmt === false) {
                throw new \Federator\Exceptions\ServerError();
            }
            $stmt->bind_param("s", $id);
            $foundId = 0;
            $ret = $stmt->bind_result($foundId);
            $stmt->execute();
            if ($ret) {
                $stmt->fetch();
            }
            $stmt->close();
        } while ($foundId > 0);

        // Add vote with created_at timestamp
        $sql = 'INSERT INTO votes (id, user_id, target_id, type, created_at) VALUES (?, ?, ?, ?, NOW())';
        $stmt = $dbh->prepare($sql);
        if ($stmt === false) {
            throw new \Federator\Exceptions\ServerError();
        }
        $stmt->bind_param("ssss", $id, $userId, $targetId, $type);
        $stmt->execute();
        $stmt->close();
        return $id; // Return the generated vote ID
    }

    /**
     * Remove a vote (like/dislike)
     *
     * @param \mysqli $dbh
     * @param string $userId
     * @param string $targetId
     * @return bool             true on success
     */
    public static function removeVote($dbh, $userId, $targetId)
    {
        $sql = 'DELETE FROM votes WHERE user_id = ? AND target_id = ?';
        $stmt = $dbh->prepare($sql);
        if ($stmt === false) {
            throw new \Federator\Exceptions\ServerError();
        }
        $stmt->bind_param("ss", $userId, $targetId);
        $stmt->execute();
        $affectedRows = $stmt->affected_rows;
        $stmt->close();
        return $affectedRows > 0;
    }
}
